struct voxelRead {
    ivec2 ID;          //shape and emission ID
    vec3 tint;
    vec2 tileIndexFraction; //2x8 bits
    vec2 tileIndexFloored;  //2x4 bits
    float tileRes;          //4 bits
};

voxelRead decodeVoxel(uvec4 bf) {
    voxelRead data;

    vec2 temp1      = unpack2x8(bf.w);

    data.ID         = unpack2x8I(bf.x);
    data.tint       = unpackColor16(bf.y);
    data.tileIndexFraction = unpack2x8(bf.z);
    data.tileIndexFloored = decode2x4Unorm(temp1.x);
    data.tileRes    = temp1.y;

    return data;
}

vec2 voxelUV(vec3 pos, vec3 normal, voxelRead voxel) {
    vec2 uv;

	if (abs(normal.y) > abs(normal.x) && abs(normal.y) > abs(normal.z)) {
		uv = vec2(fract(pos.x), fract(pos.z * sign(normal.y)));
	} else {
		uv = vec2(fract(pos.x * sign(normal.z) - pos.z * sign(normal.x)), fract(-pos.y));
	}

    uv += 256.0 * 15.0 * voxel.tileIndexFloored + 255.0 * voxel.tileIndexFraction;
    uv *= exp2(voxel.tileRes * 15.0) / textureSize(colortex3, 0);

    return uv;
}
bool alphaTestVoxel(vec3 pos, vec3 normal, voxelRead voxelData, sampler2D atlas) {
    vec2 uv     = voxelUV(pos, normal, voxelData);

    int lod     = min(1, atlasMipmapLevel);

    return texelFetch(atlas, ivec2(uv * atlasSize / (2*lod)), lod).a > 0.1;
}


ivec3 scaleIndex(ivec3 index, int targetLod) {
    vec3 x  = vec3(index) / float(targetLod);

    return ivec3(floor(x));
}

bool isVoxelOccupied(ivec3 index, int level) {
    return texture(shadowcolor1, getVoxelCoord(index, level))[level] > 0.5;
}

bool marchBVH(vec3 pos, ivec3 index, vec3 dir) {
    if (outsideVoxelVolume(index)) return false;

    ivec3 deltaSign     = ivec3(sign(dir));

    if(all(equal(deltaSign, ivec3(0)))) return false;

    ivec3 positiveDir   = max(deltaSign, ivec3(0));
    ivec3 negativeDir   = min(deltaSign, ivec3(0));

    vec3 deltaT         = 1.0 / abs(dir);

    const int maxLevel  = OCTREE_LEVELS - 1;

    vec3 steps          = (index - pos) * deltaSign + positiveDir - 1;

    int level           = 0;

    while(level++ < maxLevel && !isVoxelOccupied(index, level));

    --level;

    bool hit            = false;
    int iterations      = 0;

    do {
        float minT;
        bvec3 isMinComp;

        bool potentialHit   = false;
        int octreeIterations = 0;

        if (outsideVoxelVolume(index)) break;

        do {
            int size        = int(pow(2, level));
            ivec3 plane     = (scaleIndex(index, size) + positiveDir) * size;
            vec3 planeT     = abs(plane - pos) * deltaT;

            minT            = minOf(planeT);

            if (minT > 256.0) break;

            isMinComp       = bvec3(planeT.x == minT, planeT.y == minT, planeT.z == minT);

            ivec3 prevIndex = index;

            ivec3 stepsToNext = ivec3(minT * abs(dir) - steps + 0.5 * vec3(isMinComp));
                steps      += stepsToNext;
                index      += stepsToNext * deltaSign;

            if (outsideVoxelVolume(index)) break;

            int levelAbove  = level + 1;
            int sizeAbove   = int(pow(2, levelAbove));

            bool stepUp     = scaleIndex(prevIndex, sizeAbove) != scaleIndex(index, sizeAbove);
                level       = min(stepUp ? levelAbove : level, maxLevel);

            bool blocked    = isVoxelOccupied(index, level);

            while (blocked && level > 0) {
                blocked     = isVoxelOccupied(index, --level);
            }

            potentialHit    = blocked && level == 0;
        } while(++octreeIterations < 64 && !potentialHit);

        if (potentialHit) hit = true;

    } while (++iterations < 64 && !hit);

    return hit;
}

bool marchBVH(vec3 pos, ivec3 index, vec3 dir, out vec3 hitPos, out vec3 hitIndex, out vec3 hitNormal, out voxelRead voxel) {
    if (outsideVoxelVolume(index)) return false;

    ivec3 deltaSign     = ivec3(sign(dir));

    if(all(equal(deltaSign, ivec3(0)))) return false;

    ivec3 positiveDir   = max(deltaSign, ivec3(0));
    ivec3 negativeDir   = min(deltaSign, ivec3(0));

    vec3 deltaT         = 1.0 / abs(dir);

    const int maxLevel  = OCTREE_LEVELS - 1;

    vec3 steps          = (index - pos) * deltaSign + positiveDir - 1;

    int level           = 0;

    while(level++ < maxLevel && !isVoxelOccupied(index, level));

    --level;

    bool hit            = false;
    int iterations      = 0;

    do {
        float minT;
        bvec3 isMinComp;

        bool potentialHit   = false;
        int octreeIterations = 0;

        if (outsideVoxelVolume(index)) break;

        do {
            int size        = int(pow(2, level));
            ivec3 plane     = (scaleIndex(index, size) + positiveDir) * size;
            vec3 planeT     = abs(plane - pos) * deltaT;

            minT            = minOf(planeT);

            if (minT > 256.0) break;

            isMinComp       = bvec3(planeT.x == minT, planeT.y == minT, planeT.z == minT);

            ivec3 prevIndex = index;

            ivec3 stepsToNext = ivec3(minT * abs(dir) - steps + 0.5 * vec3(isMinComp));
                steps      += stepsToNext;
                index      += stepsToNext * deltaSign;

            if (outsideVoxelVolume(index)) break;

            int levelAbove  = level + 1;
            int sizeAbove   = int(pow(2, levelAbove));

            bool stepUp     = scaleIndex(prevIndex, sizeAbove) != scaleIndex(index, sizeAbove);
                level       = min(stepUp ? levelAbove : level, maxLevel);

            bool blocked    = isVoxelOccupied(index, level);

            while (blocked && level > 0) {
                blocked     = isVoxelOccupied(index, --level);
            }

            potentialHit    = blocked && level == 0;
        } while(++octreeIterations < 64 && !potentialHit);

        if (potentialHit) {
            vec2 voxelCoord = getVoxelCoord(index);
            hitNormal       = vec3(isMinComp) * -deltaSign;

            float hitDist   = minT;

            voxel           = decodeVoxel(uvec4(texture(shadowcolor0, voxelCoord, 0) * 65535.0));

            hit             = intersectVoxel(pos - index, dir, voxel.ID.x, hitDist, hitNormal);
            hitPos          = pos + hitDist * dir;

            #ifdef voxelAlphaTestEnabled
            if (hit) hit = alphaTestVoxel(pos + hitDist * dir, hitNormal, voxel, colortex3);
            #endif

            hitIndex        = index;
        }

    } while (++iterations < 64 && !hit);

    return hit;
}


bool marchVoxel(vec3 pos, ivec3 index, vec3 dir) {
    if (outsideVoxelVolume(index)) return false;

	vec3 nextT;
	vec3 deltaDist = 1.0 / abs(dir);
	ivec3 deltaSign;

	for (int axis = 0; axis < 3; ++axis) {
		if (dir[axis] < 0.0) {
			deltaSign[axis] = -1;
			nextT[axis] = (pos[axis] - index[axis]) * deltaDist[axis];
		} else if (dir[axis] > 0.0) {
			deltaSign[axis] = 1;
			nextT[axis] = (index[axis] - pos[axis]) * deltaDist[axis] + deltaDist[axis];
		} else {
			deltaSign[axis] = 0;
			nextT[axis] = uintBitsToFloat(0x7f800000u);// = infinity
		}
	}

	if (deltaSign.x == 0 && deltaSign.y == 0 && deltaSign.z == 0) { return false; }

	bool hit = false;
	do {
		float minComp = minOf(nextT);
		if (nextT.x == minComp) {
			index.x   += deltaSign.x;
			nextT.x += deltaDist.x;
		}
		if (nextT.y == minComp) {
			index.y   += deltaSign.y;
			nextT.y += deltaDist.y;
		}
		if (nextT.z == minComp) {
			index.z   += deltaSign.z;
			nextT.z += deltaDist.z;
		}

		if (outsideVoxelVolume(index)) break;

        vec2 voxelCoord = getVoxelCoord(index);
		float isVoxel = texture(shadowcolor1, voxelCoord, 0).x;

		hit = isVoxel > 0.5;
	} while (!hit);

    return hit;
}

//initialVoxel, 64, voxel, hitIdx, hitPos, hitNormal);
bool marchVoxel(vec3 pos, ivec3 index, vec3 dir, out vec3 hitPos, out vec3 hitIndex, out vec3 hitNormal, out voxelRead voxel) {
    if (outsideVoxelVolume(index)) return false;

	vec3 nextT;
	vec3 deltaDist = 1.0 / abs(dir);
	ivec3 deltaSign;

	for (int axis = 0; axis < 3; ++axis) {
		if (dir[axis] < 0.0) {
			deltaSign[axis] = -1;
			nextT[axis] = (pos[axis] - index[axis]) * deltaDist[axis];
		} else if (dir[axis] > 0.0) {
			deltaSign[axis] = 1;
			nextT[axis] = (index[axis] - pos[axis]) * deltaDist[axis] + deltaDist[axis];
		} else {
			deltaSign[axis] = 0;
			nextT[axis] = uintBitsToFloat(0x7f800000u);// = infinity
		}
	}

	if (deltaSign.x == 0 && deltaSign.y == 0 && deltaSign.z == 0) { return false; }

	bool hit = false;
    int i       = 0;
	do {
        float hitDist;

		float minComp = minOf(nextT);
		if (nextT.x == minComp) {
			hitDist   = nextT.x;
			hitNormal = vec3(-deltaSign.x, 0.0, 0.0);

			index.x   += deltaSign.x;
			nextT.x += deltaDist.x;
		}
		if (nextT.y == minComp) {
			hitDist   = nextT.y;
			hitNormal = vec3(0.0, -deltaSign.y, 0.0);

			index.y   += deltaSign.y;
			nextT.y += deltaDist.y;
		}
		if (nextT.z == minComp) {
			hitDist   = nextT.z;
			hitNormal = vec3(0.0, 0.0, -deltaSign.z);

			index.z   += deltaSign.z;
			nextT.z += deltaDist.z;
		}

		if (outsideVoxelVolume(index)) break;

        vec2 voxelCoord = getVoxelCoord(index);
        //ivec2 voxelPixel = ivec2(voxelCoord * voxelMapRes);

		float isVoxel = texture(shadowcolor1, voxelCoord, 0).x;
		hit = isVoxel > 0.5;

        if (hit) {
            voxel       = decodeVoxel(uvec4(texture(shadowcolor0, voxelCoord, 0) * 65535.0));

            hit         = intersectVoxel(pos - index, dir, voxel.ID.x, hitDist, hitNormal);
            hitPos      = pos + hitDist * dir;

            #ifdef voxelAlphaTestEnabled
            if (hit) hit = alphaTestVoxel(pos + hitDist * dir, hitNormal, voxel, colortex3);
            #endif

            hitIndex    = index;
        }
	} while (++i < 64 && !hit);

    return hit;
}
