/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 7 */
layout(location = 0) out vec4 result;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

uniform sampler2D colortex7;
uniform sampler2D colortex9;
uniform sampler2D colortex10;

uniform float far;

uniform vec2 pixelSize, viewSize;

/* ------ ATROUS ------ */

const float gaussKernel[2][2] = {
	{ 1.0 / 4.0, 1.0 / 8.0  },
	{ 1.0 / 8.0, 1.0 / 16.0 }
};

float computeSigmaL(sampler2D img, ivec2 pixelUV, float center) {
	float sum = center * gaussKernel[0][0];

    const int r = 1;
	for(int y = -r; y <= r; y++) {
		for(int x = -r; x <= r; x++) {
			if(x != 0 || y != 0) {
				ivec2 p     = pixelUV + ivec2(x, y);
				float variance = texelFetch(img, p, 0).a;
				float w     = gaussKernel[abs(x)][abs(y)];
				sum        += variance * w;
			}
		}
	}

	return max(sum, 1e-8);
}

vec4 filterImage(sampler2D img, sampler2D imgNormalDepth, vec2 uv, const int stepSize) {
	ivec2 pixelUV       = ivec2(uv * viewSize);

    vec4 normalDepthCenter  = texelFetch(imgNormalDepth, pixelUV, 0);
    vec4 colorCenter = texelFetch(img, pixelUV, 0);

    if (normalDepthCenter.a == 1.0) return colorCenter;

        normalDepthCenter.rgb = normalDepthCenter.rgb * 2.0 - 1.0;
        normalDepthCenter.a = sqr(normalDepthCenter.a) * far;

	float lumaCenter = getLuma(colorCenter.rgb);

    float frames    = texelFetch(colortex9, pixelUV, 0).w * maxFrames;
    float minBias   = (1.0 / max(frames, 1.0));
        minBias     = minBias + 0.2;

    float maxDelta  = mix(0.71, tau, saturate((frames - 8.0) / 24.0));
    float offset    = mix(0.06, 4e-3, saturate((frames - 8.0) / 24.0));

    float sigmaDistMul = 2.0 - (1.0 / (1.0 + normalDepthCenter.a / 32.0));

	float sigmaL = 1.0 / (1.0 * SVGF_STRICTNESS * sigmaDistMul * computeSigmaL(img, pixelUV, lumaCenter) + minBias * sigmaDistMul);

	vec3  sumColor    = colorCenter.rgb;
	float sumVariance = colorCenter.a;
    float sumWeight = 1.0;

	const int r = SVGF_RAD;
	for(int y = -r; y <= r; y++) {
		for(int x = -r; x <= r; x++) {
			ivec2 p = pixelUV + ivec2(x, y) * stepSize;

			if(x == 0 && y == 0)
				continue;

			float w = float(all(greaterThanEqual(p, ivec2(0)))
					&& all(lessThan(p, ivec2(viewSize * indirectResScale))));

			vec4 col        = texelFetch(img, p, 0);
			float lum       = getLuma(col.rgb);
			float distLum   = abs(lumaCenter - lum);
                distLum     = sqr(distLum) / max(lumaCenter, offset);
                distLum     = clamp(distLum, 0.0, maxDelta);

        vec4 normalDepth    = texelFetch(imgNormalDepth, p, 0);
            normalDepth.rgb = normalDepth.rgb * 2.0 - 1.0;
            normalDepth.a   = sqr(normalDepth.a) * far;

			float distDepth = abs(normalDepthCenter.a - normalDepth.a) * 4.0;

			w *= pow(max(0.0, dot(normalDepthCenter.rgb, normalDepth.rgb)), 16.0 * SVGF_NORMALEXP);
			w *= exp(-distDepth / sqrt(float(stepSize)) - distLum * sigmaL);

			sumColor       += col.rgb * w;
			sumVariance    += col.a * sqr(w);
			sumWeight      += w;
		}
	}
	sumColor    /= sumWeight;
	sumVariance /= sqr(sumWeight);

	return clamp16F(vec4(sumColor, sumVariance));
}

void main() {
    result          = texelFetch(colortex7, ivec2(coord * viewSize), 0);

    vec2 lowresCoord    = coord / indirectResScale;
    if (saturate(lowresCoord) != lowresCoord) discard;

    #ifdef SVGF_POSTACCUM
    result          = filterImage(colortex7, colortex10, coord, SVGF_SIZE);
    #endif
}