#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

layout (triangles) in;

#ifdef generateShadowmap
    layout (triangle_strip, max_vertices = 6) out;
#else
    layout (triangle_strip, max_vertices = 3) out;
#endif

uniform sampler2D gcolor;

uniform vec3 cameraPosition;

uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection;

flat in int[] vertBlockID;

in vec3[] vertTint;

/* VOXEL DATA INPUT */
flat in int[] canBeVoxelized;
flat in ivec2[] vertObjectID;

in vec2[] vertCoord;
in vec2[] vertMidCoord;

in vec3[] centerDist;

flat in vec3[] vertNormal;
flat in vec4[] vertTangent;

/* VOXEL DATA OUTPUT */
flat out int isVoxel;
flat out int matID;
flat out int octreeLevel;
flat out uvec4 voxelWrite;

/* SHADOWMAP DATA OUTPUT */
out float warp;

out vec2 coord;
out vec3 tint;

out vec3 scenePos;
out vec3 worldPos;

#include "/lib/light/warp.glsl"


/* VOXEL STUFF */
#include "/lib/voxel/store.glsl"

vec3 getQuadCenter() {
    vec3 m;
        
        m.xy    = (vertMidCoord[0] - vertCoord[2]) * inverse(
            mat2(vertCoord[0].x - vertCoord[2].x, vertCoord[1].x - vertCoord[2].x,
                 vertCoord[0].y - vertCoord[2].y, vertCoord[1].y - vertCoord[2].y
                ));
        m.z     = 1.0 - m.x - m.y;

    return mat3(gl_in[0].gl_Position.xyz, gl_in[1].gl_Position.xyz, gl_in[2].gl_Position.xyz) * m;
}

struct voxelData {
    ivec2 voxelID;
    vec3 tint;
};

struct atlasData {
    vec2 tileIndexFraction; //2x8 bits
    vec2 tileIndexFloored;  //2x4 bits
    float tileRes;          //4 bits
};

uvec4 packVoxelData0(voxelData data, atlasData atlas) {
    uvec4 bf;
        bf.x    = pack2x8UI(data.voxelID);
        bf.y    = packColor16UI(data.tint);
        bf.z    = pack2x8UI(atlas.tileIndexFraction);
        bf.w    = pack2x8UI(vec2(encode2x4Unorm(atlas.tileIndexFloored), atlas.tileRes));

    return bf;
}

void main() {
    #ifdef generateShadowmap
    for (int i = 0; i < 3; ++i) {
        isVoxel = 0;
        tint    = vertTint[i];
        coord   = vertCoord[i];
        matID   = vertBlockID[i];

        gl_Position     = gl_ModelViewMatrix * gl_in[i].gl_Position;

        scenePos        = transMAD(shadowModelViewInverse, gl_Position.xyz);
        worldPos        = cameraPosition + scenePos;

        gl_Position     = gl_ProjectionMatrix * gl_Position;
        gl_Position.xy  = shadowmapWarp(gl_Position.xy, warp);
        gl_Position.z  *= 0.2;
        gl_Position.x   = gl_Position.x * 0.5 + 0.5;
        gl_Position.xy  = gl_Position.yx;
        if (gl_Position.y <= 0.0) return;

        EmitVertex();
    } EndPrimitive();
    #endif

    if (canBeVoxelized[0] == 0) return;

    vec3 bitangent  = cross(vertTangent[0].xyz, vertNormal[0]) * (vertTangent[0].w < 0.0 ? -1.0 : 1.0);
    vec3 vertPosU   = vertTangent[0].xyz * mat3(gl_in[0].gl_Position.xyz, gl_in[1].gl_Position.xyz, gl_in[2].gl_Position.xyz);
    vec3 vertPosV   = bitangent * mat3(gl_in[0].gl_Position.xyz, gl_in[1].gl_Position.xyz, gl_in[2].gl_Position.xyz);

    vec2 faceSize   = vec2(maxOf(vertPosU) - minOf(vertPosU), maxOf(vertPosV) - minOf(vertPosV));
        faceSize   /= vec2(length(vertTangent[0].xyz), length(bitangent));

    vec2 minCoord   = min(min(vertCoord[0], vertCoord[1]), vertCoord[2]);
    vec2 maxCoord   = max(max(vertCoord[0], vertCoord[1]), vertCoord[2]);
    vec2 tileSize   = abs(maxCoord - minCoord) / faceSize;
    vec2 tileRes    = round(tileSize * textureSize(gcolor, 0));

    ivec2 tileIndex = ivec2(vertMidCoord[0] / tileSize);

    voxelData voxelTemp     = voxelData(vertObjectID[0], vertTint[0]);
    atlasData atlasTemp     = atlasData(fract(tileIndex / 256.0) * 256.0 / 255.0,
                                        floor(tileIndex / 256.0) / 15.0,
                                        log2(maxOf(tileRes)) / 15.0);

    uvec4 voxelWriteTemp    = packVoxelData0(voxelTemp, atlasTemp);

    vec3 normal     = mat3(shadowModelViewInverse) * normalize(gl_NormalMatrix * vertNormal[0]);

    float normalBias = 0.5 - 0.25 * normal.y;

    const vec2[3] verts = vec2[3](vec2(1,-1) / shadowMapRes, vec2(0,1) / shadowMapRes, vec2(-1,-1) / shadowMapRes);

    for (int j = 0; j < 3; ++j) {
        vec3 pos        = gl_in[j].gl_Position.xyz;
            pos         = transMAD(gl_ModelViewMatrix, pos);
            pos         = transMAD(shadowModelViewInverse, pos);

        vec3 voxelIndex = floor(sceneToVoxelSpace(pos + centerDist[j] * 0.9));

        if (!withinVoxelVolume(voxelIndex)) return;

        vec2 texelCoord     = getVoxelCoord(voxelIndex) * 2.0 - 1.0;

        voxelWrite  = voxelWriteTemp;
        isVoxel = 1;

        tint    = vec3(1.0);
        coord   = vec2(0.0);

        gl_Position     = vec4(texelCoord, normalBias, 1.0);
        gl_Position.xy += verts[j];

        EmitVertex();
    }
    EndPrimitive();
}