#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/


/* RENDERTARGETS: 0,4, */
layout(location = 0) out vec4 sceneColor;
layout(location = 1) out uvec4 screenLightingStore;
layout(location = 2) out vec4 translucency;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

flat in mat4x3 lightColor;

uniform sampler2D colortex0;
uniform usampler2D colortex1;
uniform usampler2D colortex2;
uniform sampler2D colortex3;
uniform usampler2D colortex4;
uniform sampler2D colortex7;
uniform sampler2D colortex10;
uniform sampler2D colortex11;
uniform sampler2D colortex14;
uniform sampler2D colortex15;

uniform sampler2D depthtex0;
uniform sampler2D depthtex2;

uniform sampler2D noisetex;

uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform int frameCounter;

uniform float aspectRatio;
uniform float lightFlip;
uniform float sunAngle;
uniform float near, far;

uniform vec2 pixelSize, viewSize;

uniform mat4 gbufferProjection;

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/frag/bluenoise.glsl"

/* ------ ATROUS ------ */

#include "/lib/offset/gauss.glsl"

#define atrousDepthTreshold 0.001
#define atrousDepthOffset 0.25
#define atrousDepthExp 2.0

#define atrousNormalExponent 16

#define atrousLumaExp vec2(halfPi, euler)

ivec2 clampTexelPos(ivec2 pos) {
    return clamp(pos, ivec2(0.0), ivec2(viewSize));
}

mat2x3 unpackDoubleRGBE8(uvec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}


vec4 packReflectionAux(vec3 directLight, vec3 albedo) {
    vec4 lightRGBE  = encodeRGBE8(directLight);
    vec4 albedoRGBE = encodeRGBE8(albedo);

    return vec4(pack2x8(lightRGBE.xy),
                pack2x8(lightRGBE.zw),
                pack2x8(albedoRGBE.xy),
                pack2x8(albedoRGBE.zw));
}

mat2x3 fetchLighting(usampler2D tex, ivec2 coord) {
    return unpackDoubleRGBE8(texelFetch(tex, coord, 0));
}


mat2 computeVariance(usampler2D tex, ivec2 pos) {
    vec2 sumMsqr    = vec2(0.0);
    vec2 sumMean    = vec2(0.0);

    for (int i = 0; i<9; i++) {
        ivec2 deltaPos     = kernelO_3x3[i];

        mat2x3 col = fetchLighting(tex, clampTexelPos(pos + deltaPos));
        vec2 lum   = vec2(avgOf(col[0]), avgOf(col[1]));

        sumMsqr    += sqr(lum);
        sumMean    += lum;
    }
    sumMsqr  /= 9.0;
    sumMean  /= 9.0;

    vec2 variance  = abs(sumMsqr - sqr(sumMean)) * rcp(max(sumMean, 1e-20));

    return mat2(variance.x, sumMean.x, variance.y, sumMean.y);
}

vec3 spatialUpscale(sampler2D tex, sampler2D gData, vec3 sceneNormal, float centerDepth) {
    ivec2 pos           = ivec2(coord * viewSize * indirectResScale);

    vec4 centerData     = vec4(sceneNormal, sqrt(depthLinear(centerDepth)) * far);

    vec3 centerColor    = texelFetch(tex, pos, 0).rgb;

    vec3 total          = vec3(0);
    float totalWeight   = 0.0;

    for (int i = 0; i<25; i++) {
        ivec2 deltaPos      = kernelO_5x5[i];

        ivec2 samplePos     = pos + deltaPos;

        bool valid          = all(greaterThanEqual(samplePos, ivec2(0))) && all(lessThan(samplePos, ivec2(viewSize)));

        if (!valid) continue;

        vec4 currentData    = texelFetch(gData, clampTexelPos(samplePos), 0);
            currentData.xyz = currentData.xyz * 2.0 - 1.0;
            currentData.w   = (currentData.w) * far;

        float depthDelta    = abs(currentData.w - centerData.w) * atrousDepthExp;

        float spatialWeight = pow(max(0.0, dot(currentData.xyz, centerData.xyz)), atrousNormalExponent);
            spatialWeight  *= exp(-depthDelta);

        //if (weight < 1e-20) continue;

        vec3 currentColor   = texelFetch(tex, samplePos, 0).rgb;

        //accumulate stuff
        total       += currentColor * spatialWeight;

        totalWeight += spatialWeight;
    }

    //compensate for total sampling weight
    total          /= max(totalWeight, 1e-16);

    return total;
}


float rgbHue(vec3 rgb) {
    float hue;
    if (rgb.x == rgb.y && rgb.y == rgb.z) hue = 0.0;
    else hue = (180.0 * rcp(pi)) * atan(2.0 * rgb.x - rgb.y - rgb.z, sqrt(3.0) * (rgb.y - rgb.z));

    if (hue < 0.0) hue = hue + 360.0;

    return clamp(hue, 0.0, 360.0);
}
float rgbSaturation(vec3 rgb) {
    float minrgb    = minOf(rgb);
    float maxrgb    = maxOf(rgb);

    return (max(maxrgb, 1e-10) - max(minrgb, 1e-10)) / max(maxrgb, 1e-10);
}

uvec4 packScreenLighting(vec3 directLight, vec3 emission) {
    vec4 lightRGBE  = encodeRGBE8(directLight);
    vec4 emissionRGBE = encodeRGBE8(emission);

    return uvec4(pack2x8UI(lightRGBE.xy),
                pack2x8UI(lightRGBE.zw),
                pack2x8UI(emissionRGBE.xy),
                pack2x8UI(emissionRGBE.zw));
}

vec2 offsetDist(float x) {
	float n = fract(x * 8.0) * pi;
    return vec2(cos(n), sin(n)) * x;
}

float getSSAO(sampler2D depthtex, float depth, vec2 coord, float dither) {
    const uint steps = 4;
    const float baseRadius = sqrt2;

    bool hand       = depth < 0.56;
        depth       = depthLinear(depth);

    float currStep  = 0.2 * dither + 0.2;
	float fovScale  = gbufferProjection[1][1] / 1.37;
	float distScale = max((far - near) * depth + near, 5.0);
	vec2 scale      = baseRadius * vec2(1.0 / aspectRatio, 1.0) * fovScale / distScale;

    float ao = 0.0;

    const float maxOcclusionDist    = tau;
    const float anibleedExp         = 0.71;

    for (uint i = 0; i < steps; ++i) {
		vec2 offset = offsetDist(currStep) * scale;
		float mult  = (0.7 / baseRadius) * (far - near) * (hand ? 1024.0 : 1.0);

		float sampleDepth = depthLinear(texture(depthtex, coord + offset).r);
		float sample0 = (depth - sampleDepth) * mult;
        float antiBleed = 1.0 - rcp(1.0 + max0(distance(sampleDepth, depth) * far - maxOcclusionDist) * anibleedExp);
            antiBleed   = saturate(antiBleed);
		float angle = mix(clamp(0.5 - sample0, 0.0, 1.0), 0.5, antiBleed);
		float dist  = mix(clamp(0.25 * sample0 - 1.0, 0.0, 1.0), 0.5, antiBleed);

		sampleDepth = depthLinear(texture(depthtex, coord - offset).r);
		sample0     = (depth - sampleDepth) * mult;
        antiBleed   = 1.0 - rcp(1.0 + max0(distance(sampleDepth, depth) * far - maxOcclusionDist) * anibleedExp);
           antiBleed   = saturate(antiBleed);
        angle      += mix(clamp(0.5 - sample0, 0.0, 1.0), 0.5, antiBleed);
        dist       += mix(clamp(0.25 * sample0 - 1.0, 0.0, 1.0), 0.5, antiBleed);
		
		ao         += (clamp(angle + dist, 0.0, 1.0));
		currStep   += 0.2;
    }
	ao *= 1.0 / float(steps);

    if (saturate(ao) != ao) ao = 1.0;
	
	return sqrt(saturate(ao));
}


/*

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;

uniform vec3 cameraPosition;

uniform vec2 taaOffset;
*/

//#include "/lib/util/transforms.glsl"

//#include "/lib/voxel/store.glsl"
//#include "/lib/voxel/octree.glsl"
//#include "/lib/voxel/intersect.glsl"
//#include "/lib/voxel/trace.glsl"

/*
vec3[8] octreeColor = vec3[8] (
    vec3(1.0, 0.0, 0.0),
    vec3(1.0, 1.0, 0.0),
    vec3(0.0, 1.0, 0.0),
    vec3(0.0, 0.0, 1.0),
    vec3(0.0, 0.0, 1.0),
    vec3(0.0, 1.0, 0.5),
    vec3(0.0, 1.0, 1.0),
    vec3(0.0, 0.5, 1.0)
);

vec4 octreeVis = vec4(0);
vec4 octreeVis2 = vec4(0);



bool isVoxelOccupied(ivec3 index, int level) {
    octreeVis += vec4(octreeColor[level], 1.0);

    if(texture(shadowcolor1, getVoxelCoord(index, level))[level] > 0.5) octreeVis2 += vec4(octreeColor[level], 1.0);

    return texture(shadowcolor1, getVoxelCoord(index, level))[level] > 0.5;
}

bool voxelRaytrace2(vec3 pos, ivec3 index, vec3 dir, inout vec3 octreeVis, inout int steps, int level) {
    int size    = int(pow(2, level));

    if (outsideVoxelVolume(index)) return false;

    if (isVoxelOccupied(index, level)) return true;

    index   = ivec3(floor(vec3(index) / float(size)));
    pos    /= size;
    //index *= size;
    //size = 1;

    ivec3 startIndex = index;

	vec3 nextT;
	vec3 deltaDist = 1.0 / abs(dir);
	ivec3 deltaSign;

	for (int axis = 0; axis < 3; ++axis) {
		if (dir[axis] < 0.0) {
			deltaSign[axis] = -1;
			nextT[axis] = (pos[axis] - index[axis]) * deltaDist[axis];
		} else if (dir[axis] > 0.0) {
			deltaSign[axis] = 1;
			nextT[axis] = (index[axis] - pos[axis]) * deltaDist[axis] + deltaDist[axis];
		} else {
			deltaSign[axis] = 0;
			nextT[axis] = uintBitsToFloat(0x7f800000u);// = infinity
		}
	}

	if (deltaSign.x == 0 && deltaSign.y == 0 && deltaSign.z == 0) { return false; }

	bool hit = false;
	do {
		float minComp = minOf(nextT);
		if (nextT.x == minComp) {
			index.x    += deltaSign.x;
			nextT.x    += deltaDist.x;
		}
		if (nextT.y == minComp) {
			index.y    += deltaSign.y;
			nextT.y    += deltaDist.y;
		}
		if (nextT.z == minComp) {
			index.z    += deltaSign.z;
			nextT.z    += deltaDist.z;
		}

		if (outsideVoxelVolume(index * size)) break;

        hit     = isVoxelOccupied(index * size, level);

        if (hit) octreeVis += octreeColor[level];

        steps++;
	} while (!hit);

    return hit;
}


vec3 voxelDebug(float sceneDepth) {
    vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth), true);
    vec3 scenePos   = viewToSceneSpace(viewPos);

    vec3 cameraPos  = sceneToVoxelSpace(gbufferModelViewInverse[3].xyz);

    vec3 voxelPos   = sceneToVoxelSpace(scenePos);
    vec3 voxelIndex = floor(cameraPos);
    vec3 startIndex = floor(cameraPos);

    vec3 direction  = normalize(voxelPos - cameraPos);

    vec3 vis   = vec3(0.0);
    int steps   = 1;

    //bool hit    = voxelRaytrace2(cameraPos, ivec3(voxelIndex), direction, vis, steps, 0);
    //    hit    = voxelRaytrace2(cameraPos, ivec3(voxelIndex), direction, vis, steps, 1);
    //    hit    = voxelRaytrace2(cameraPos, ivec3(voxelIndex), direction, vis, steps, 2);
    //    hit    = voxelRaytrace2(cameraPos, ivec3(voxelIndex), direction, vis, steps, 3);
    bool hit    = marchBVH(cameraPos, ivec3(voxelIndex), direction);

    return octreeVis2.rgb / 32;
}

*/

void main() {
    sceneColor          = stex(colortex0);
    float sceneDepth    = stex(depthtex0).x;
    vec3 albedo         = vec3(1.0);
    vec3 directLight    = vec3(1.0);
    //indirectLuma        = 0.0;
    //temporalAux         = stex(colortex14);

    if (landMask(sceneDepth)) {
        uvec4 tex1      = texelFetch(colortex1, ivec2(coord * viewSize), 0);
        vec3 sceneNormal = decodeNormal(invF16UI(tex1.xy));

        bool hand   = sceneDepth < stex(depthtex2).x;

            albedo          = sceneColor.rgb;
            
        mat2x3 screenLighting = unpackDoubleRGBE8(texelFetch(colortex4, ivec2(coord * viewSize), 0));

        //mat2x3 ptLight      = unpackDoubleRGBE8(texelFetch(colortex7, ivec2(coord * viewSize * indirectResScale), 0));
        vec3 ptLight        = spatialUpscale(colortex7, colortex10, sceneNormal, sceneDepth).rgb;

        float ssao          = getSSAO(depthtex0, sceneDepth, coord, ditherBluenoise());

        vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * lightFlip;
        screenLighting[0]  *= directCol;

        vec3 emission       = screenLighting[1] / pi;
        if (hand) emission /= sqrPi;

        sceneColor.rgb     *= screenLighting[0] + ptLight * ssao + emission + minimumAmbientColor * minimumAmbientMult * ssao * minimumAmbientIllum;

        //sceneColor.rgb      = ptLight;

        #if DEBUG_VIEW == 3
            sceneColor.rgb      = ptLight;
        #endif
        //indirectLuma        = avgOf(indirectLight);

        //float emissionHue   = rgbHue(ptLight[0]);
        //temporalAux.y       = emissionHue / 360.0;
        //temporalAux.z       = rgbSaturation(ptLight[0]);
        //temporalAux.w       = getLuma(ptLight[0]) / 128.0;

        //sceneColor.rgb  = vec3(texture(colortex11, coord * indirectResScale));
        screenLightingStore   = packScreenLighting(clamp16F(screenLighting[0]), clamp16F(albedo));
    }

    //sceneColor.rgb  = albedo*0 + voxelDebug(sceneDepth) * 4;

    sceneColor          = drawbufferClamp(sceneColor);
    translucency        = vec4(0.0);
}